/* zklCurl.c : Front end to libCurl: http://curl.haxx.se/libcurl/
 * 
 * Libraries you will need: libcurl
 * 
 * libcurl is licensed under a MIT/X derivate license:
 *    http://curl.haxx.se/docs/copyright.html
 * 
 * Copyright (c) 2014 Craig Durland craigd@zenkinetic.com
 * License: The zlib/libpng License,
 * 	http://www.opensource.org/licenses/zlib-license.php
 */

#include "zklObject.h"
#include "zklData.h"
#include "zklList.h"
#include "zklString.h"
#include "zklUtil.h"
#include "zklImports.h"

#include <string.h>
#include <curl/curl.h>

static ZKL_Object CurlObject;

typedef struct
{
   Instance  instance;	// Inherit from Instance
   CURL     *curl;
   unsigned  open:1, writingBody:2, nowrite:1;
   Instance *in;	  // input to cURL, to be sent
   Instance *out;	  // data read by cURL
   size_t    headerSize;  // bytes of http header read by cURL
   size_t    trailerSize; // bytes of http header read by cURL
   pVM	     vm;	  // only valid during perform
} ZCurl;

static void curlMarker(Instance *_zc)
{
   ZCurl *zc = (ZCurl *)_zc;
   instanceMark(zc->in);
   instanceMark(zc->out);
}

    // CURLOPT_READFUNCTION: called by libcurl when it wants data
size_t _reader(void *ptr, size_t size,size_t nmemb, void *zpc)
{
   ZCurl *zc     = (ZCurl *)zpc;
   Instance *in  = zc->in;
   size_t wsz    = size*nmemb;
   size_t cursor = dataCursorPos(in), dsz = dataSize(in) - cursor;

   if (!dsz) return 0;
   if (wsz < dsz) wsz = dsz;
   memcpy(ptr,dataChunkAt(in,cursor,wsz),wsz);

   dataCursorSet(in,cursor + wsz);

   return wsz;
}

static size_t _writer(Byte *ptr, size_t size, size_t nmemb, ZCurl *zc)
{
   size_t numBytes = size*nmemb;
   if(zc->nowrite) return numBytes;
//!!! fence and close on error
   if (!zc->out)	// first write
   {
      zc->out = dataCreate(1048575,I_OWNED,zc->vm);	// 1mb
      dataInsertBytes(zc->out,0,ptr,numBytes,0,zc->vm);
      return numBytes;
   }
   return dataInsertBytes(zc->out, 0, ptr,numBytes, 1,zc->vm); // bytes written
}

    // CURLOPT_WRITEFUNCTION: called by libcurl to write data to me
    // Either: writing header, body or trailer (which is a header after body)
    // I think trailers are only for [broken] HTTP streams
static size_t _writeHeader(Byte *ptr, size_t size, size_t nmemb, void *zpc)
{
   ZCurl *zc = (ZCurl *)zpc;
#if 0
   if (zc->writingBody) zc->trailerSize += (size*nmemb);
   else 		zc->headerSize  += (size*nmemb);
#else
   if (zc->writingBody==0x2) return nmemb; // ignore first header
   if (zc->writingBody) zc->trailerSize += (size*nmemb);
#endif
   else 		zc->headerSize  += (size*nmemb);
   return _writer(ptr,size,nmemb,zc);
}
static size_t _writeBody(Byte *ptr, size_t size, size_t nmemb, void *zpc)
{
   ZCurl *zc = (ZCurl *)zpc;
   zc->writingBody = 0x1;
   return _writer(ptr,size,nmemb,zc);
}

    // CURLOPT_SEEKFUNCTION: aka fseek on input stream
    // Returns:
    //   CURL_SEEKFUNC_OK on success, 
    //   CURL_SEEKFUNC_FAIL to cause the upload operation to fail 
    //   CURL_SEEKFUNC_CANTSEEK to indicate that while the seek failed,
    //      libcurl is free to work around the problem if possible
    // origin: SEEK_SET, SEEK_CUR. SEEK_END
int _seek(void *instream, curl_off_t offset, int origin)
{
   ZCurl *zc = (ZCurl *)instream;
   Instance *in = zc->in;
   size_t cursor = dataCursorPos(in), dsz = dataSize(in);
   curl_off_t n = 0;

   switch(origin)
   {
      case SEEK_CUR: n = cursor; break;	// cursor
      case SEEK_SET: n = 0; 	 break;	// start
      case SEEK_END: n = dsz;	 break;	// end
   }
   n += offset;

   dataCursorSet(in,(size_t)n);
   return CURL_SEEKFUNC_OK;
}

/* ******************************************************************** */
/* ***************************** Methods ****************************** */
/* ******************************************************************** */

    // .create()
static Instance *curl_create(Instance *x,pArglist arglist,pVM vm)
{
   ZCurl *zc = (ZCurl *)instanceAllocate(sizeof(ZCurl),&CurlObject,1,vm);
   CURL  *curl;
   CURLcode  r;

   zc->in = zc->out = 0;
   curl = curl_easy_init();
   if (!curl) vmThrow(vm,E_IO_ERROR,"libcurl doesn't want to init");
   zc->curl = curl;
   
   r = curl_easy_setopt(curl,CURLOPT_READFUNCTION,_reader);
   r = curl_easy_setopt(curl,CURLOPT_READDATA,zc);

   r = curl_easy_setopt(curl,CURLOPT_HEADERFUNCTION,_writeHeader);
   r = curl_easy_setopt(curl,CURLOPT_HEADERDATA,zc);
   r = curl_easy_setopt(curl,CURLOPT_WRITEFUNCTION,_writeBody);
   r = curl_easy_setopt(curl,CURLOPT_WRITEDATA,zc);

   r = curl_easy_setopt(curl,CURLOPT_SEEKFUNCTION,_seek);
   r = curl_easy_setopt(curl,CURLOPT_SEEKDATA,zc);

   zc->open = 1; zc->nowrite = 0;
   return addToCollectables((Instance *)zc,I_OWNED,vm);
}

static int curlClose(Instance *_zc)	// also GC freeme routine
{
   ZCurl *zc = (ZCurl *)_zc;
   if (zc->open)
   {
      curl_easy_cleanup(zc->curl);	// turn off writing, etc
      zc->open = 0;
      zc->in = zc->out = 0;
   }
   return 1;
}
    // .close()
static Instance *curl_close(ZCurl *self,pArglist arglist,pVM vm)
   { curlClose((Instance *)self); return Void; }

static int nm2opt(char *name,Instance *pp,pVM vm)
{
   typedef struct{ char *name; int v,t; } Nm2Opt;
   static Nm2Opt opts[] =
   {
      "VERBOSE",	CURLOPT_VERBOSE,	IntType,
      "HEADER",		CURLOPT_HEADER,		IntType,
      "WILDCARDMATCH",	CURLOPT_WILDCARDMATCH,	IntType,
//      "PATH_AS_IS",	CURLOPT_PATH_AS_IS,	IntType,
      "PROXY",		CURLOPT_PROXY,		StringType,
      "PROXYPORT",	CURLOPT_PROXYPORT,	IntType,
      "NOPROXY",	CURLOPT_NOPROXY,	StringType,

      "USERPWD",	CURLOPT_USERPWD,	StringType,
      "_USERNAME",	CURLOPT_USERNAME,	StringType,
      "PASSWORD",	CURLOPT_PASSWORD,	StringType,

      "SSLCERT",	CURLOPT_SSLCERT,	StringType,
      "SSLCERTTYPE",	CURLOPT_SSLCERTTYPE,	StringType,
      "SSLKEY",		CURLOPT_SSLKEY,		StringType,
      0,		0,			0,
   };
   Nm2Opt *ptr;
   int     t = TYPEO(pp);
   if (t==BoolType) t = IntType;
   for(ptr=opts; ptr->name; ptr++)
      if (0==strcmp(name,ptr->name))
      {
	 if(t != ptr->t) vmThrow(vm,E_TYPE_ERROR,"CURL.setOpt value");
	 return ptr->v;
      }
   vmThrow(vm,E_VALUE_ERROR,"CURL.setOpt: unknown option");
   return 0;
}


    // .setOpt("option name",value[Bool|Int|String]) -->self
static Instance *curl_setOpt(ZCurl *self,pArglist arglist,pVM vm)
{
   CURLcode   s;
   Instance  *pp=arglistGet(arglist,1,"CURL.setOpt",vm);
   int        copt;

   copt = nm2opt(arglistGetOnlyString(arglist,0,"",vm),pp,vm);
   switch(TYPEO(pp))
   {
      case IntType:    s = curl_easy_setopt(self->curl,copt,(long)INT64V(pp)); break;
      case BoolType:   s = curl_easy_setopt(self->curl,copt,(long)BOOLV(pp));  break;
      case StringType: s = curl_easy_setopt(self->curl,copt,stringText(pp));   break;
      default: vmThrow(vm,E_TYPE_ERROR,"CURL.setOpt");
   }
   if (s)
   {
      char buf[200];
      sprintf(buf, "curl_easy_setopt() failed: %s",curl_easy_strerror(s));
      vmThrow(vm,E_IO_ERROR,buf);
   }

   return (Instance *)self;
}

#if 0
    // .perform() -->self
static Instance *curl_perform(ZCurl *self,pArglist arglist,pVM vm)
{
   CURLcode  r;
   if (self->open)
   {
      self->vm = vm;
         r = curl_easy_perform(self->curl);
      self->vm = 0;
      if (r != CURLE_OK)
      {
	 char buf[200];
	 sprintf(buf, "curl_easy_perform() failed: %s",curl_easy_strerror(r));
	 vmThrow(vm,E_IO_ERROR,buf);
      }
   }
   return (Instance *)self;
}
#endif


    // .get(url,headerOnly=False|Void)
    //  -->(Data, bytes of header, bytes of trailer)
static Instance *curl_get(ZCurl *self,pArglist arglist,pVM vm)
{
   CURL     *curl = self->curl;
   char     *url  = arglistGetOnlyString(arglist,0,"curl.get([url])",vm);
//   long	     headerOnly = arglistTryToGetBool(arglist,1,0,0,vm);
   Instance *headerOnly = arglistTryToGet(arglist,1);
   CURLcode  r;

   if (!self->open) return Void;
//??? in this case, create, get, close?

   if(self->out) dataWillFill(self->out);	// reuse buffer
   self->headerSize = self->trailerSize = 0;
   self->writingBody = self->nowrite = 0;
//      curl_easy_reset(curl);
   r = curl_easy_setopt(curl,CURLOPT_URL,url);
   
   r = curl_easy_setopt(curl,CURLOPT_FOLLOWLOCATION,1L);

#if 0
   curl_easy_setopt(curl,CURLOPT_NOBODY,headerOnly);
//   curl_easy_setopt(curl,CURLOPT_HEADER,headerOnly);
#else
   if(headerOnly==Void)	// don't know why CURLOPT_HEADER=0 don't work
   {  
      curl_easy_setopt(curl,CURLOPT_NOBODY,0);
      self->writingBody = 0x2;	// ignore header writes
   }
   else curl_easy_setopt(curl,CURLOPT_NOBODY,headerOnly==BoolTrue);
#endif
   vmMightStall(vm,1); self->vm = vm;
      r = curl_easy_perform(self->curl);
   self->vm = 0; vmMightStall(vm,0);
   if (r != CURLE_OK)
   {
      char buf[200];
      sprintf(buf, "curl_easy_perform() failed: %s",curl_easy_strerror(r));
      vmThrow(vm,E_IO_ERROR,buf);
   }
//   curl_easy_reset(curl);
   return tupleCreateX(vm, self->out,
      intCreate(self->headerSize,vm),intCreate(self->trailerSize,vm),
      ZNIL);
}

#if 0
static Instance *curl_pingURL(ZCurl *self,pArglist arglist,pVM vm)
{
   CURL     *curl = self->curl;
   char     *url  = arglistGetOnlyString(arglist,0,"curl.get([url])",vm);
   CURLcode  r;

   if (!self->open) return Void;

   self->out = 0; self->headerSize = self->trailerSize = 0;
   self->writingBody = 0;
   self->nowrite = 1;
//      curl_easy_reset(curl);
   r = curl_easy_setopt(curl,CURLOPT_URL,url);
   
   r = curl_easy_setopt(curl,CURLOPT_FOLLOWLOCATION,1L);
   curl_easy_setopt(curl,CURLOPT_NOBODY,1L);
   vmMightStall(vm,1); self->vm = vm;
      r = curl_easy_perform(self->curl);
   self->vm = 0; vmMightStall(vm,0);
   if (r != CURLE_OK)
   {
      char buf[200];
      sprintf(buf, "curl_easy_perform() failed: %s",curl_easy_strerror(r));
      vmThrow(vm,E_IO_ERROR,buf);
   }
//   curl_easy_reset(curl);
   return self->headerSize ? BoolTrue : BoolFalse;
}
#endif

    // .urlEncode(string) -->string
static Instance *curl_urlEncode(ZCurl *self,pArglist arglist,pVM vm)
{
   char *url = arglistGetOnlyString(arglist,0,"Curl.urlEncode",vm);
   char *ptr = curl_easy_escape(self->curl,url,0);
   if (ptr)
   {
      Instance *r = stringCreate(ptr,I_OWNED,vm);
      curl_free(ptr);
      return r;
   }
   return emptyString;
}

static Instance *curl_urlDecode(ZCurl *self,pArglist arglist,pVM vm)
{
   char *url = arglistGetOnlyString(arglist,0,"Curl.urlDecode",vm);
//!!! %00 is valid, which might mean use Data
   char *ptr = curl_easy_unescape(self->curl,url,0,0);
   if (ptr)
   {
      Instance *r = stringCreate(ptr,I_OWNED,vm);
      curl_free(ptr);
      return r;
   }
   return emptyString;
}

static Instance *curl_verbose(ZCurl *self,pArglist arglist,pVM vm)
{
   int b = arglistGetBool(arglist,0,"Curl.verbose",vm);
   curl_easy_setopt(self->curl,CURLOPT_VERBOSE,(long)b);
   return (Instance *)self;
}

static MethodTable methods[] =
{
   "create",	(pMethod)curl_create,
   "close",	(pMethod)curl_close,
   "setOpt",	(pMethod)curl_setOpt,
//   "perform",	(pMethod)curl_perform,
   "get",	(pMethod)curl_get,
//   "pingURL",	(pMethod)curl_pingURL,
   "urlEncode", (pMethod)curl_urlEncode,
   "urlDecode", (pMethod)curl_urlDecode,
   "verbose",	(pMethod)curl_verbose,
   0,		0
};


/* ******************************************************************** */
/* **************************** Properties **************************** */
/* ******************************************************************** */

    // .version -->string
static Instance *version(Instance *self, pVM vm)
   { return kStringCreate(curl_version( ), 0,I_OWNED,vm); }

    // .authors
static Instance *authors(Instance *self, pVM vm)
{
   return createTupleOfStrings(vm,"Craig Durland","http://zenkinetic.com/",
          (char *)0); 
}

    // .libraries
static Instance *libraries(Instance *self, pVM vm)
{
   return createTupleOfStrings(vm,
      "libcurl","",
      (char *)0);
}

static PropertyTable properties[] =
{
   "version",		(pProperty)version,
   "authors",		(pProperty)authors,
   "libraries",		(pProperty)libraries,
   0,			0
};


DllExport void *construct(void *vm)
{
   Instance *zc;

   if (!vm) return (void *)ZKLX_PROTOCOL;	// handshake
   	// If this is reloaded, nothing happens except 
	// construct() is called again so don't reinitialize
   if (!CurlObject.name)	// static items are zero
   {
      CURLcode r = curl_global_init(CURL_GLOBAL_DEFAULT);
      if (r)
      {
	 char buf[200];
	 sprintf(buf, "curl_easy_perform() failed: %s",curl_easy_strerror(r));
	 vmThrow(vm,E_LOADER_ERROR,buf);
      }

      constructObject(&CurlObject,NativeType, methods,properties,0,vm);
      CurlObject.name        = "Curl";
      CurlObject.vaultPath   = "Network";
      CurlObject.isize       = sizeof(ZCurl);
      CurlObject.freeMe      = curlClose;
      CurlObject.magicMarker = curlMarker;
   }
   zc = curl_create(Void,NoArglist,vm);
   curlClose(zc);
   return zc;
}
